/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung - Bibliothek
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.lib;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import javax.swing.event.EventListenerList;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Beobachter für die Änderung von Sollparametern.<br>
 * Die Klasse vereinheitlicht und zentralisiert die Anmeldung für
 * Parameterdatensätze. Es erfolgt eine zentrale Anmeldung für den Aspekt
 * "asp.parameterSoll". Die angemeldeten Listener werden bei einer Änderung mit
 * den entsprechenden Daten versorgt.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class ParameterBeobachter implements ClientReceiverInterface {

	/**
	 * Klasse zur Speicherung der Daten, die einem Parameterlistener beim
	 * anmelden sofort gemeldet werden können.
	 *
	 * @author peuker
	 */
	private static class ParameterListenerCache extends EventListenerList {

		/** Versions-ID zur Serialisierung. */
		private static final long serialVersionUID = 1L;

		/** die Daten des Eintrags. */
		private ResultData daten;

		/** Konstruktor. */
		ParameterListenerCache() {
			super();
		}

		/**
		 * liefert die Daten des Eintrags.
		 *
		 * @return die Daten
		 */
		ResultData getDaten() {
			return daten;
		}

		/**
		 * setzt die Daten des Eintrags.
		 *
		 * @param daten
		 *            die Daten
		 */
		void setDaten(final ResultData daten) {
			this.daten = daten;
		}
	}

	/**
	 * Thread zur Entkopplung der Benachrichtigung der angemeldeten Listener, um
	 * Deadlocks bei der Kommunikation mit dem Datenverteiler zu vermeiden.
	 *
	 * @author BitCtrl Systems GmbH, Peuker
	 */
	private static class ParameterNotifier extends Thread {
		/**
		 * Klasse zur Speicherung der Daten einer Listener-Benachrichtigung
		 * inklusie des zu benachrichtigenden Listeners..
		 *
		 * @author BitCtrl Systems GmbH, Peuker
		 */
		private static class NotifierElement {

			/** der Listener. */
			private final ParameterListener listener;

			/** die an den Listener zu übertragenden Daten. */
			private final ResultData result;

			/**
			 * Konstruktor.
			 *
			 * @param listener
			 *            der Listener
			 * @param result
			 *            die Daten
			 */
			NotifierElement(final ParameterListener listener, final ResultData result) {
				this.listener = listener;
				this.result = result;
			}

			/**
			 * liefert den gespeicherten Listener.
			 *
			 * @return den Listener
			 */
			ParameterListener getListener() {
				return listener;
			}

			/**
			 * liefert die gespeicherten Daten.
			 *
			 * @return die Daten
			 */
			ResultData getResult() {
				return result;
			}

		}

		/** Logger für Debug-Ausgaben. */
		private static final Debug NOTIFIERLOGGER = Debug.getLogger();

		/** die Liste der auszuführenden Benachrichtigungen. */
		private final LinkedList<NotifierElement> queue = new LinkedList<>();

		/**
		 * Konstruktor.
		 */
		ParameterNotifier() {
			super("ParameterNotifier");
		}

		/**
		 * fügt der Liste der Benachrichtigungen eine neue hinzu.
		 *
		 * @param listener
		 *            der zu benachrichtigende Listener
		 * @param result
		 *            die Daten
		 */
		void add(final ParameterListener listener, final ResultData result) {
			synchronized (queue) {
				queue.add(new NotifierElement(listener, result));
			}
		}

		/** {@inheritDoc} */
		@Override
		public void run() {
			while (true) {
				synchronized (queue) {
					while (!queue.isEmpty()) {
						final NotifierElement element = queue.getFirst();
						queue.removeFirst();
						element.getListener().neuerParameter(element.getResult());
					}
				}
				synchronized (this) {
					try {
						this.wait(MethodenBibliothek.MILLIS_PER_MINUTE);
					} catch (final InterruptedException e) {
						ParameterNotifier.NOTIFIERLOGGER.warning("ParameterNotifier wurde unterbrochen: ",
								new Object[] { e });
					}
				}
			}
		}
	}

	/** Logger für Debug-Ausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/** die verwendete Datenverteilerverbindung. */
	private final ClientDavInterface verbindung;

	/** der Aspekt für den Empfang von Soll-Parametern. */
	private final Aspect sollAsp;

	/**
	 * die Listener für die jeweiligen Datenspezifikationen für
	 * Parameter-Datensätze.
	 */
	private final Map<ParameterInfo, ParameterListenerCache> parameterList = new HashMap<>();

	/**
	 * der Thread der die Listener im Hintergrund benachrichtigt.
	 */
	private final ParameterNotifier notifier;

	/**
	 * konstruiert den Beobachter für die übergebene Datenverteilerverbindung.
	 *
	 * @param verbindung
	 *            die verwendete Datenverteilerverbindung
	 */
	ParameterBeobachter(final ClientDavInterface verbindung) {
		this.verbindung = verbindung;
		final DataModel model = verbindung.getDataModel();
		sollAsp = model.getAspect("asp.parameterSoll");
		notifier = new ParameterNotifier();
		notifier.start();
	}

	/**
	 * meldet den Empfang der durch die übergebene Spezifikation definierten
	 * Parameter ab.
	 *
	 * @param info
	 *            die Parameterspezifikation
	 */
	private void abmelden(final ParameterInfo info) {
		verbindung.unsubscribeReceiver(this, info.getObjekt(),
				new DataDescription(info.getAtg(), sollAsp, info.getSim()));
	}

	/**
	 * installiert den übergebenen Listener als {@link ParameterListener} für
	 * die übergebenen Parameterspezifikationen.
	 *
	 * @param listener
	 *            der Listener
	 * @param parameter
	 *            dte Spezifikationen
	 * @throws ParameterClientException
	 *             die übergebenen Parameterspezifikationen enthielten ungültige
	 *             Einträge.
	 */
	void addListener(final ParameterListener listener, final ParameterInfo... parameter)
			throws ParameterClientException {
		pruefeAnmeldeDaten(parameter);
		for (final ParameterInfo info : parameter) {
			ParameterInfo usedInfo = info;
			if (info.getSim() == DataDescription.NO_SIMULATION_VARIANT_SET) {
				final short sim = verbindung.getClientDavParameters().getSimulationVariant();
				usedInfo = new ParameterInfo(info.getObjekt(), info.getTyp(), info.getAtg(), sim);
			}
			ParameterListenerCache liste = parameterList.get(usedInfo);
			if (null == liste) {
				liste = new ParameterListenerCache();
				parameterList.put(usedInfo, liste);
				liste.add(ParameterListener.class, listener);
				anmelden(usedInfo);
			} else {
				liste.add(ParameterListener.class, listener);
				if (null != liste.getDaten()) {
					listener.neuerParameter(liste.getDaten());
				}
			}
		}
	}

	/**
	 * meldet sich für den Empfang der durch die Spezifikation definierten
	 * Parameter an.
	 *
	 * @param info
	 *            die Spezifikation des Parameters
	 */
	private void anmelden(final ParameterInfo info) {
		short sim = info.getSim();
		if (sim == DataDescription.NO_SIMULATION_VARIANT_SET) {
			sim = verbindung.getClientDavParameters().getSimulationVariant();
		}

		verbindung.subscribeReceiver(this, info.getObjekt(), new DataDescription(info.getAtg(), sollAsp, sim),
				ReceiveOptions.normal(), ReceiverRole.receiver());
	}

	/**
	 * prüft die gültigkeit der übergebenen Parameterspezifikationen.
	 *
	 * @param parameter
	 *            die Menge der Spezifikationen
	 * @throws ParameterClientException
	 *             es wurden ungültige Spezifikationen übergeben
	 */
	private void pruefeAnmeldeDaten(final ParameterInfo[] parameter) throws ParameterClientException {
		final Collection<ParameterInfo> fehlerListe = new ArrayList<>();
		for (final ParameterInfo info : parameter) {
			if (!info.getObjekt().getType().equals(info.getTyp())) {
				fehlerListe.add(info);
			}
		}

		if (fehlerListe.size() > 0) {
			throw new ParameterClientException(
					"Ungültige Parameterspezifikationen für die Anmeldung als ParameterListener ermittelt",
					fehlerListe);
		}
	}

	/**
	 * entfernt den angegebenen {@link ParameterListener} als Listener für die
	 * angegebenen Parameterspezifikationen.
	 *
	 * @param listener
	 *            der Listener
	 * @param parameter
	 *            die Menge der abzumeldenden Spezifikationen
	 */
	void removeListener(final ParameterListener listener, final ParameterInfo... parameter) {
		for (final ParameterInfo info : parameter) {
			final EventListenerList liste = parameterList.get(info);
			if (null != liste) {
				liste.remove(ParameterListener.class, listener);
				if (liste.getListenerCount() <= 0) {
					parameterList.remove(info);
					abmelden(info);
				}
			}
		}
	}

	/**
	 * {@inheritDoc}
	 *
	 * Die Funktion empfängt die Daten für angemeldete Parameterdatensätze und
	 * verteilt die Daten an die angemeldeten {@link ParameterListener}.
	 *
	 * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
	 */
	@Override
	public void update(final ResultData[] results) {

		ParameterBeobachter.LOGGER.finest("Parameteraktualisierungen empfangen");

		for (final ResultData result : results) {
			ParameterBeobachter.LOGGER.finest("Neuer Parameterdatensatz: ", result);
			final ParameterInfo info = new ParameterInfo(result.getObject(),
					result.getDataDescription().getAttributeGroup(),
					result.getDataDescription().getSimulationVariant());
			final ParameterListenerCache liste = parameterList.get(info);
			if (null != liste) {
				liste.setDaten(result);
				for (final ParameterListener listener : liste.getListeners(ParameterListener.class)) {
					notifier.add(listener, result);
					// listener.neuerParameter(result);
				}
			}
		}
		synchronized (notifier) {
			notifier.notifyAll();
		}
	}
}
